#ifndef _SEPOL_IBENDPORT_INTERNAL_H_
#define _SEPOL_IBENDPORT_INTERNAL_H_

#include <sepol/ibendport_record.h>
#include <sepol/ibendports.h>
#include "dso.h"

hidden_proto(sepol_ibendport_create)
hidden_proto(sepol_ibendport_free)
hidden_proto(sepol_ibendport_get_con)
hidden_proto(sepol_ibendport_get_port)
hidden_proto(sepol_ibendport_key_create)
hidden_proto(sepol_ibendport_key_unpack)
hidden_proto(sepol_ibendport_set_con)
hidden_proto(sepol_ibendport_set_port)
hidden_proto(sepol_ibendport_get_ibdev_name)
hidden_proto(sepol_ibendport_set_ibdev_name)
#endif
